"""An untyped serializer"""

import json
from typing import Any, Type

from ..config import SerializerConfig


def _serialize_key_if_str(key: Any, config: SerializerConfig) -> Any:
    return config.serialize_key(
        key
    ) if config.serialize_key and isinstance(key, str) else key


def _from_value(
        value: Any,
        type_annotation: Type,
        config: SerializerConfig
) -> Any:
    serializer = config.value_serializers.get(type_annotation)
    if serializer is not None:
        return serializer(value)
    return value


def _from_list(lst: list, config: SerializerConfig) -> list:
    return [
        _from_any(item, config)
        for item in lst
    ]


def _from_dict(dct: dict, config: SerializerConfig) -> dict:
    return {
        _serialize_key_if_str(key, config): _from_any(value, config)
        for key, value in dct.items()
    }


def _from_any(obj: Any, config: SerializerConfig) -> Any:
    if isinstance(obj, dict):
        return _from_dict(obj, config)
    elif isinstance(obj, list):
        return _from_list(obj, config)
    else:
        return _from_value(obj, type(obj), config)


def serialize(obj: Any, config: SerializerConfig) -> str:
    json_obj = _from_any(obj, config)
    return json.dumps(
        json_obj,
        indent=2 if config.pretty_print else None
    )



"""Serialization"""

from typing import Any

import jetblack_serialization.typing_inspect_ex as typing_inspect

from ..types import (
    Annotation,
)
from ..config import SerializerConfig


from .typed_serializer import serialize as typed_serialize
from .typed_deserializer import deserialize as typed_deserialize
from .untyped_serializer import serialize as untyped_serialize
from .untyped_deserializer import deserialize as untyped_deserialize


def _is_typed(annotation: Annotation) -> bool:
    return (
        typing_inspect.is_typed_dict_type(annotation) or
        (
            typing_inspect.is_list_type(annotation) and
            _is_typed(typing_inspect.get_args(annotation)[0])
        ) or
        (
            typing_inspect.is_annotated_type(annotation) and
            _is_typed(typing_inspect.get_origin(annotation))
        )
    )


def serialize(
        obj: Any,
        annotation: Any,
        config: SerializerConfig
) -> str:
    """Convert the object to JSON

    Args:
        obj (Any): The object to convert
        annotation (Annotation): The type annotation
        config (SerializerConfig): The serializer configuration

    Returns:
        str: The serialized object
    """
    if _is_typed(annotation):
        return typed_serialize(obj, annotation, config)
    else:
        return untyped_serialize(obj, config)


def deserialize(
        text: str,
        annotation: Annotation,
        config: SerializerConfig,
) -> Any:
    """Convert JSON to an object

    Args:
        text (str): The JSON string
        annotation (Annotation): The type annotation
        config (SerializerConfig): The serializer configuration

    Returns:
        Any: The deserialized object.
    """
    if _is_typed(annotation):
        return typed_deserialize(text, annotation, config)
    else:
        return untyped_deserialize(text, config)



"""An XML serializer"""

from decimal import Decimal
from inspect import Parameter
import json
from typing import Any, Type, Union, cast

import jetblack_serialization.typing_inspect_ex as typing_inspect
from ..config import SerializerConfig
from ..types import Annotation
from ..utils import is_simple_type

from .annotations import (
    JSONAnnotation,
    JSONValue,
    JSONProperty,
    is_json_annotation,
    get_json_annotation
)


def _from_value(
        value: Any,
        type_annotation: Type,
        config: SerializerConfig
) -> Any:
    if type_annotation is str:
        return value
    elif type_annotation is int:
        return value
    elif type_annotation is bool:
        return value
    elif type_annotation is float:
        return value
    elif type_annotation is Decimal:
        return float(value)
    else:
        serializer = config.value_serializers.get(type_annotation)
        if serializer is not None:
            return serializer(value)

    raise TypeError(f'Unhandled type {type_annotation}')


def _from_optional(
        obj: Any,
        type_annotation: Annotation,
        json_annotation: JSONAnnotation,
        config: SerializerConfig
) -> Any:
    if obj is None:
        return None

    # An optional is a union where the last element is the None type.
    union_types = typing_inspect.get_args(type_annotation)[:-1]
    if len(union_types) == 1:
        # This was Optional[T]
        return _from_any(
            obj,
            union_types[0],
            json_annotation,
            config
        )
    else:
        return _from_union(
            obj,
            Union[tuple(union_types)],
            json_annotation,
            config
        )


def _from_union(
        obj: Any,
        type_annotation: Annotation,
        json_annotation: JSONAnnotation,
        config: SerializerConfig
) -> Any:
    for element_type in typing_inspect.get_args(type_annotation):
        try:
            return _from_any(
                obj,
                element_type,
                json_annotation,
                config
            )
        except:  # pylint: disable=bare-except
            pass


def _from_list(
        lst: list,
        type_annotation: Annotation,
        config: SerializerConfig
) -> Any:
    item_annotation, *_rest = typing_inspect.get_args(type_annotation)
    if typing_inspect.is_annotated_type(item_annotation):
        item_type_annotation, item_json_annotation = get_json_annotation(
            item_annotation
        )
    else:
        item_type_annotation = item_annotation
        item_json_annotation = JSONValue()

    return [
        _from_any(
            item,
            item_type_annotation,
            item_json_annotation,
            config
        )
        for item in lst
    ]


def _from_typed_dict(
        dct: dict,
        type_annotation: Annotation,
        config: SerializerConfig
) -> dict:
    json_obj = dict()

    typed_dict_keys = typing_inspect.typed_dict_keys(type_annotation)
    for key, key_annotation in typed_dict_keys.items():
        default = getattr(type_annotation, key, Parameter.empty)
        if typing_inspect.is_annotated_type(key_annotation):
            item_type_annotation, item_json_annotation = get_json_annotation(
                key_annotation
            )
            if not issubclass(type(item_json_annotation), JSONProperty):
                raise TypeError("<ust be a property")
            json_property = cast(JSONProperty, item_json_annotation)
        else:
            property_name = config.serialize_key(
                key
            ) if isinstance(key, str) else key
            json_property = JSONProperty(property_name)
            item_type_annotation = key_annotation

        value = dct.get(key, default)
        if value != Parameter.empty:
            json_obj[json_property.tag] = _from_any(
                value,
                item_type_annotation,
                json_property,
                config
            )
        else:
            # TODO: Should we throw here?
            pass

    return json_obj


def _from_any(
        value: Any,
        type_annotation: Annotation,
        json_annotation: JSONAnnotation,
        config: SerializerConfig
) -> Any:
    if is_simple_type(type_annotation):
        return _from_value(
            value,
            type_annotation,
            config
        )
    elif typing_inspect.is_optional_type(type_annotation):
        return _from_optional(
            value,
            type_annotation,
            json_annotation,
            config
        )
    elif typing_inspect.is_list_type(type_annotation):
        return _from_list(
            value,
            type_annotation,
            config
        )
    elif typing_inspect.is_typed_dict_type(type_annotation):
        return _from_typed_dict(
            value,
            type_annotation,
            config
        )
    elif typing_inspect.is_union_type(type_annotation):
        return _from_union(
            value,
            type_annotation,
            json_annotation,
            config
        )
    else:
        raise TypeError('Unhandled type')


def serialize(
        obj: Any,
        annotation: Annotation,
        config: SerializerConfig
) -> str:
    """Serialize an object to JSON

    Args:
        obj (Any): The object to serialize
        annotation (Annotation): The objects type annotation

    Raises:
        TypeError: If the object cannot be serialized

    Returns:
        str: The JSON string
    """
    if is_json_annotation(annotation):
        type_annotation, json_annotation = get_json_annotation(annotation)
    else:
        type_annotation, json_annotation = annotation, JSONValue()

    json_obj = _from_any(
        obj,
        type_annotation,
        json_annotation,
        config
    )
    return json.dumps(
        json_obj,
        indent=2 if config.pretty_print else None
    )



"""XML Serialization"""

from decimal import Decimal
from inspect import Parameter
from typing import (
    Any,
    Dict,
    Iterable,
    List,
    Optional,
    Type,
    Union
)

from lxml import etree
from lxml.etree import _Element  # pylint: disable=no-name-in-module

from .. import typing_inspect_ex as typing_inspect
from ..config import SerializerConfig
from ..custom_annotations import get_typed_dict_key_default
from ..types import Annotation
from ..utils import is_simple_type

from .annotations import (
    XMLAnnotation,
    XMLAttribute,
    XMLEntity,
    get_xml_annotation,
    DEFAULT_DOCUMENT_ANNOTATION,
    DEFAULT_KEY_ANNOTATION,
    DEFAULT_VALUE_ANNOTATION,
)

def _is_element_empty(element: _Element, xml_annotation: XMLAnnotation) -> bool:
    if isinstance(xml_annotation, XMLAttribute):
        return xml_annotation.tag not in element.attrib
    else:
        return (
            element.find('*') is None and
            element.text is None and
            not element.attrib
        )


def _to_value(
        text: Optional[str],
        default: Any,
        type_annotation: Type,
        config: SerializerConfig
) -> Any:
    if text is None:
        return default
    elif type_annotation is str:
        return text
    elif type_annotation is int:
        return int(text)
    elif type_annotation is bool:
        return text.lower() == 'true'
    elif type_annotation is float:
        return float(text)
    elif type_annotation is Decimal:
        return Decimal(text)
    else:
        deserializer = config.value_deserializers.get(type_annotation)
        if deserializer is not None:
            return deserializer(text)

    raise TypeError(f'Unhandled type {type_annotation}')


def _to_union(
        element: Optional[_Element],
        type_annotation: Annotation,
        xml_annotation: XMLAnnotation,
        config: SerializerConfig
) -> Any:
    for union_type_annotation in typing_inspect.get_args(type_annotation):
        try:
            return _to_obj(
                element,
                Parameter.empty,
                union_type_annotation,
                xml_annotation,
                config
            )
        except:  # pylint: disable=bare-except
            pass
    raise ValueError('Unable to deserialize a Union')


def _to_optional(
        element: Optional[_Element],
        type_annotation: Annotation,
        xml_annotation: XMLAnnotation,
        config: SerializerConfig
) -> Any:
    if element is None or _is_element_empty(element, xml_annotation):
        return None

    # An optional is a union where the last element is the None type.
    union_types = typing_inspect.get_args(type_annotation)[:-1]
    if len(union_types) == 1:
        # This was Optional[T]
        return _to_obj(
            element,
            Parameter.empty,
            union_types[0],
            xml_annotation,
            config
        )
    else:
        return _to_union(
            element,
            Union[tuple(union_types)], # type: ignore
            xml_annotation,
            config
        )


def _to_simple(
        element: Optional[_Element],
        default: Any,
        type_annotation: Annotation,
        xml_annotation: XMLAnnotation,
        config: SerializerConfig
) -> Any:
    if element is None:
        raise ValueError('Found "None" while deserializing a value')
    if not isinstance(xml_annotation, XMLAttribute):
        text = element.text
    else:
        text = element.attrib[xml_annotation.tag]
    if text is None and default is Parameter.empty:
        raise ValueError(f'Expected "{xml_annotation.tag}" to be non-null')
    if isinstance(text, bytes):
        text = text.decode()
    return _to_value(text, default, type_annotation, config)


def _to_list(
        element: Optional[_Element],
        type_annotation: Annotation,
        xml_annotation: XMLAnnotation,
        config: SerializerConfig
) -> List[Any]:
    if element is None:
        raise ValueError('Received "None" while deserializing a list')

    item_annotation, *_rest = typing_inspect.get_args(type_annotation)
    if typing_inspect.is_annotated_type(item_annotation):
        item_type_annotation, item_xml_annotation = get_xml_annotation(
            item_annotation,
            DEFAULT_VALUE_ANNOTATION
        )
    else:
        item_type_annotation = item_annotation
        item_xml_annotation = xml_annotation

    if xml_annotation.tag == item_xml_annotation.tag:
        # siblings
        elements: Iterable[_Element] = element.iterfind(
            '../' + item_xml_annotation.tag)
    else:
        # nested
        elements = element.iter(item_xml_annotation.tag)

    return [
        _to_obj(
            child,
            Parameter.empty,
            item_type_annotation,
            item_xml_annotation,
            config
        )
        for child in elements
    ]


def _to_typed_dict(
        element: Optional[_Element],
        type_annotation: Annotation,
        config: SerializerConfig
) -> Optional[Dict[str, Any]]:
    if element is None:
        raise ValueError('Received "None" while deserializing a TypeDict')

    typed_dict: Dict[str, Any] = {}

    typed_dict_keys = typing_inspect.typed_dict_keys(type_annotation)
    for key, key_annotation in typed_dict_keys.items():
        default = get_typed_dict_key_default(key_annotation)
        if typing_inspect.is_annotated_type(key_annotation):
            item_type_annotation, item_xml_annotation = get_xml_annotation(
                key_annotation,
                DEFAULT_KEY_ANNOTATION
            )
        else:
            tag = config.serialize_key(key) if isinstance(key, str) else key
            item_xml_annotation = XMLEntity(tag)
            item_type_annotation = typing_inspect.get_unannotated_type(
                key_annotation
            )

        if isinstance(item_xml_annotation, XMLAttribute) or item_xml_annotation.tag == '':
            item_element = element
        else:
            item_element = element.find('./' + item_xml_annotation.tag)

        typed_dict[key] = _to_obj(
            item_element,
            default,
            item_type_annotation,
            item_xml_annotation,
            config
        )

    return typed_dict


def _to_obj(
        element: Optional[_Element],
        default: Optional[Any],
        type_annotation: Annotation,
        xml_annotation: XMLAnnotation,
        config: SerializerConfig
) -> Any:

    if is_simple_type(type_annotation):
        return _to_simple(
            element,
            default,
            type_annotation,
            xml_annotation,
            config
        )
    if typing_inspect.is_optional_type(type_annotation):
        return _to_optional(
            element,
            type_annotation,
            xml_annotation,
            config
        )
    elif typing_inspect.is_list_type(type_annotation):
        return _to_list(
            element,
            type_annotation,
            xml_annotation,
            config
        )
    elif typing_inspect.is_typed_dict_type(type_annotation):
        return _to_typed_dict(
            element,
            type_annotation,
            config)
    elif typing_inspect.is_union_type(type_annotation):
        return _to_union(
            element,
            type_annotation,
            xml_annotation,
            config
        )
    raise TypeError


def deserialize(
        text: str,
        annotation: Annotation,
        config: SerializerConfig
) -> Any:
    """Convert XML to an object

    Args:
        text (str): The XML string
        annotation (str): The type annotation

    Returns:
        Any: The deserialized object.
    """
    type_annotation, xml_annotation = get_xml_annotation(
        annotation,
        DEFAULT_DOCUMENT_ANNOTATION
        )
    if not isinstance(xml_annotation, XMLEntity):
        raise TypeError(
            "Expected the root value to have an XMLEntity annotation")

    element = etree.fromstring(text)  # pylint: disable=c-extension-no-member
    return _to_obj(element, Parameter.empty, type_annotation, xml_annotation, config)



"""XML Serialization"""

from .annotations import (
    XMLAttribute,
    XMLEntity
)

from .deserializer import deserialize
from .serializer import serialize

__all__ = [
    'XMLAttribute',
    'XMLEntity',

    'deserialize',
    'serialize'
]



"""JSON Serialization"""

from .annotations import JSONValue, JSONProperty
from .typed_deserializer import (
    from_json_value
)
from .serialization import (
    serialize,
    deserialize
)

__all__ = [
    'JSONValue',
    'JSONProperty',

    'serialize',
    'deserialize',
    'from_json_value'
]



"""Serializer Config"""

from datetime import datetime, timedelta
from typing import Any, Callable, Dict, Optional, Type

from jetblack_iso8601 import (
    iso8601_to_datetime,
    iso8601_to_timedelta,
    datetime_to_iso8601,
    timedelta_to_iso8601,
)


def _same_name(name: str) -> str:
    return name


def to_datetime(text: str) -> datetime:
    value = iso8601_to_datetime(text)
    if value is None:
        raise ValueError('Unable to parse iso8601 timestamp')
    return value


def to_timedelta(text: str) -> timedelta:
    value = iso8601_to_timedelta(text)
    if value is None:
        raise ValueError('Unable to parse iso8601 timestamp')
    return value


VALUE_DESERIALIZERS: Dict[Type, Callable[[str], Any]] = {
    datetime: to_datetime,
    timedelta: to_timedelta
}
VALUE_SERIALIZERS: Dict[Type, Callable[[Any], str]] = {
    datetime: datetime_to_iso8601,
    timedelta: timedelta_to_iso8601
}


class SerializerConfig:
    """Configuration for serialization"""

    def __init__(
        self,
        serialize_key: Optional[Callable[[str], str]],
        deserialize_key: Optional[Callable[[str], str]],
        *,
        pretty_print: bool = False,
        value_serializers=VALUE_SERIALIZERS,
        value_deserializers=VALUE_DESERIALIZERS
    ) -> None:
        self.serialize_key = serialize_key or _same_name
        self.deserialize_key = deserialize_key or _same_name
        self.pretty_print = pretty_print
        self.value_serializers = value_serializers
        self.value_deserializers = value_deserializers



"""Protocol utilities"""

from datetime import datetime, timedelta
from decimal import Decimal
from typing import Any

import jetblack_serialization.typing_inspect_ex as typing_inspect
from .types import Annotation


def is_simple_type(annotation: Annotation) -> bool:
    """Return True if the annotation is a simple type like an int or a str.

    Args:
        annotation (Any): The annotation

    Returns:
        bool: True if the annotation is a JSON literal, otherwise False
    """
    return annotation in (
        str,
        bool,
        int,
        float,
        Decimal,
        datetime,
        timedelta
    )


def is_container_type(annotation: Any) -> bool:
    """Return True if this is a JSON container.

    A JSON container can be an object (Like a Dict[str, Any]), or a List.

    Args:
        annotation (Any): The type annotation.

    Returns:
        bool: True if the annotation is represented in JSON as a container.
    """
    if typing_inspect.is_optional_type(annotation):
        return is_container_type(typing_inspect.get_optional_type(annotation))
    else:
        return (
            typing_inspect.is_list_type(annotation) or
            typing_inspect.is_dict_type(annotation) or
            typing_inspect.is_typed_dict_type(annotation)
        )



"""Custom annotations"""

from abc import ABCMeta
from inspect import Signature
from typing import Any, List, Tuple, Type, TypeVar

import jetblack_serialization.typing_inspect_ex as typing_inspect
from .types import Annotation

T = TypeVar('T')


class SerializationAnnotation(metaclass=ABCMeta):
    """The base serialization annotation class"""


class DefaultValue:

    def __init__(self, value: Any) -> None:
        self.value = value


def is_any_annotation_of_type(annotation: Annotation, tp: Type[Any]) -> bool:
    if not typing_inspect.is_annotated_type(annotation):
        return False
    for item in typing_inspect.get_metadata(annotation):
        if issubclass(type(item), tp):
            return True
    return False


def get_all_annotations_of_type(
        annotation: Annotation,
        tp: Type[T]
) -> Tuple[Annotation, List[T]]:
    type_annotation = typing_inspect.get_origin(annotation)
    serialization_annotations = [
        serialization_annotation
        for serialization_annotation in typing_inspect.get_metadata(annotation) or []
        if issubclass(type(serialization_annotation), tp)
    ]
    return type_annotation, serialization_annotations


def is_any_serialization_annotation(annotation: Annotation) -> bool:
    """Determine if the annotation is of type Annotation[T, SerializationAnnotation]

    Args:
        annotation (Any): The annotation

    Returns:
        bool: True if the annotation is of type
            Annotation[T, SerializationAnnotation], otherwise False
    """
    return is_any_annotation_of_type(annotation, SerializationAnnotation)


def get_all_serialization_annotations(
        annotation: Annotation
) -> Tuple[Annotation, List[SerializationAnnotation]]:
    """Gets the type T of Annotation[T, SerializationAnnotation]

    Args:
        annotation (Any): The annotation

    Returns:
        Tuple[Annotation, List[SerializationAnnotation]]: The type and the
            serialization annotationa
    """
    return get_all_annotations_of_type(annotation, SerializationAnnotation)


def is_any_default_annotation(annotation: Annotation) -> bool:
    return is_any_annotation_of_type(annotation, DefaultValue)


def get_default_annotation(
        annotation: Annotation
) -> Tuple[Annotation, DefaultValue]:
    typ, annotations = get_all_annotations_of_type(
        annotation, DefaultValue)
    assert len(annotations) == 1, "There can be only one"
    return typ, annotations[0]


def get_typed_dict_key_default(td):
    if is_any_default_annotation(td):
        _, annotation = get_default_annotation(td)
        return annotation.value
    return Signature.empty



"""JSON Serialization"""

from decimal import Decimal
from inspect import Parameter
import json
from typing import (
    Any,
    Dict,
    List,
    Optional,
    Type,
    Union,
    cast
)

from .. import typing_inspect_ex as typing_inspect
from ..config import SerializerConfig
from ..custom_annotations import get_typed_dict_key_default
from ..types import Annotation
from ..utils import is_simple_type

from .annotations import (
    JSONAnnotation,
    JSONValue,
    JSONProperty,
    is_json_annotation,
    get_json_annotation
)


def _to_value(
        value: Any,
        type_annotation: Type,
        config: SerializerConfig
) -> Any:
    if isinstance(value, type_annotation):
        return value

    if isinstance(value, str):
        if type_annotation is str:
            return value
        elif type_annotation is int:
            return int(value)
        elif type_annotation is bool:
            return value.lower() == 'true'
        elif type_annotation is float:
            return float(value)
        elif type_annotation is Decimal:
            return Decimal(value)
        else:
            deserializer = config.value_deserializers.get(type_annotation)
            if deserializer is not None:
                return deserializer(value)
    elif isinstance(value, (int, float)) and type_annotation is Decimal:
        return Decimal(value)

    raise TypeError(f'Unhandled type {type_annotation}')


def _to_optional(
        obj: Any,
        type_annotation: Annotation,
        json_annotation: JSONAnnotation,
        config: SerializerConfig
) -> Any:
    # An optional is a union where the last element is the None type.
    union_types = typing_inspect.get_args(type_annotation)[:-1]
    if len(union_types) == 1:
        return None if not obj else _to_any(
            obj,
            union_types[0],
            json_annotation,
            config
        )
    else:
        union = Union[tuple(union_types)]  # type: ignore
        return _to_any(
            obj,
            union,
            json_annotation,
            config
        )


def _to_list(
        lst: list,
        list_annotation: Annotation,
        config: SerializerConfig
) -> List[Any]:
    item_type_annotation, *_rest = typing_inspect.get_args(list_annotation)
    if typing_inspect.is_annotated_type(item_type_annotation):
        item_type_annotation, item_json_annotation = get_json_annotation(
            item_type_annotation
        )
    else:
        item_json_annotation = JSONValue()

    return [
        _to_any(
            item,
            item_type_annotation,
            item_json_annotation,
            config
        )
        for item in lst
    ]


def _to_union(
        obj: Optional[Dict[str, Any]],
        type_annotation: Annotation,
        json_annotation: JSONAnnotation,
        config: SerializerConfig
) -> Any:
    for item_type_annotation in typing_inspect.get_args(type_annotation):
        try:
            return _to_any(
                obj,
                item_type_annotation,
                json_annotation,
                config
            )
        except:  # pylint: disable=bare-except
            pass


def _to_dict(
        obj: Dict[str, Any],
        dict_annotation: Annotation,
        config: SerializerConfig
) -> Dict[str, Any]:
    json_obj: Dict[str, Any] = {}

    typed_dict_keys = typing_inspect.typed_dict_keys(dict_annotation)
    for key, key_annotation in typed_dict_keys.items():
        default = get_typed_dict_key_default(key_annotation)
        if is_json_annotation(key_annotation):
            item_type_annotation, item_json_annotation = get_json_annotation(
                key_annotation
            )
            if not issubclass(type(item_json_annotation), JSONProperty):
                raise TypeError("Must be a property")
            json_property = cast(JSONProperty, item_json_annotation)
        else:
            tag = config.serialize_key(
                key
            ) if isinstance(key, str) else key
            json_property = JSONProperty(tag)
            item_type_annotation = typing_inspect.get_unannotated_type(
                key_annotation
            )

        if json_property.tag in obj:
            json_obj[key] = _to_any(
                obj[json_property.tag],
                item_type_annotation,
                json_property,
                config
            )
        elif default != Parameter.empty:
            json_obj[key] = _to_any(
                default,
                item_type_annotation,
                json_property,
                config
            )
        elif typing_inspect.is_optional_type(item_type_annotation):
            json_obj[key] = None
        else:
            raise KeyError(
                f'Required key "{json_property.tag}" is missing'
            )

    return json_obj


def _to_any(
        json_value: Any,
        type_annotation: Annotation,
        json_annotation: JSONAnnotation,
        config: SerializerConfig
) -> Any:
    if is_simple_type(type_annotation):
        return _to_value(
            json_value,
            type_annotation,
            config
        )
    elif typing_inspect.is_optional_type(type_annotation):
        return _to_optional(
            json_value,
            type_annotation,
            json_annotation,
            config
        )
    elif typing_inspect.is_list_type(type_annotation):
        return _to_list(
            json_value,
            type_annotation,
            config
        )
    elif typing_inspect.is_typed_dict_type(type_annotation):
        return _to_dict(
            json_value,
            type_annotation,
            config
        )
    elif typing_inspect.is_union_type(type_annotation):
        return _to_union(
            json_value,
            type_annotation,
            json_annotation,
            config
        )
    else:
        raise TypeError


def from_json_value(
        config: SerializerConfig,
        json_value: Any,
        annotation: Annotation,
) -> Any:
    """Convert from a json value

    Args:
        config (SerializerConfig): The serialier configuration
        json_value (Any): The JSON value
        annotation (Annotation): The type annotation

    Raises:
        TypeError: If the value cannot be deserialized to the type

    Returns:
        Any: The deserialized value
    """
    if is_json_annotation(annotation):
        type_annotation, json_annotation = get_json_annotation(annotation)
        if not isinstance(json_annotation, JSONValue):
            raise TypeError(
                "Expected the root value to have a JSONValue annotation"
            )
    else:
        type_annotation, json_annotation = annotation, JSONValue()

    return _to_any(
        json_value,
        type_annotation,
        json_annotation,
        config
    )


def deserialize(
        text: str,
        annotation: Annotation,
        config: SerializerConfig
) -> Any:
    """Convert JSON to an object

    Args:
        text (str): The JSON string
        annotation (str): The type annotation

    Returns:
        Any: The deserialized object.
    """
    json_value = json.loads(text)
    return from_json_value(config, json_value, annotation)



"""An XML serializer"""

from decimal import Decimal
from inspect import Parameter
from typing import Any, Optional, Type, Union

from lxml import etree
from lxml.etree import Element, _Element, SubElement  # pylint: disable=no-name-in-module

import jetblack_serialization.typing_inspect_ex as typing_inspect
from ..types import Annotation
from ..config import SerializerConfig
from ..utils import is_simple_type

from .annotations import (
    XMLAnnotation,
    XMLAttribute,
    XMLEntity,
    get_xml_annotation,
    DEFAULT_DOCUMENT_ANNOTATION,
    DEFAULT_KEY_ANNOTATION,
    DEFAULT_VALUE_ANNOTATION
)


def _make_element(parent: Optional[_Element], tag: str) -> _Element:
    return Element(tag) if parent is None else SubElement(parent, tag)


def _from_value(
        value: Any,
        type_annotation: Type,
        config: SerializerConfig
) -> str:
    if type_annotation is str:
        return value
    elif type_annotation is int:
        return str(value)
    elif type_annotation is bool:
        return 'true' if value else 'false'
    elif type_annotation is float:
        return str(value)
    elif type_annotation is Decimal:
        return str(value)
    else:
        serializer = config.value_serializers.get(type_annotation)
        if serializer is not None:
            return serializer(value)

    raise TypeError(f'Unhandled type {type_annotation}')


def _from_optional(
        obj: Any,
        type_annotation: Annotation,
        xml_annotation: XMLAnnotation,
        element: Optional[_Element],
        config: SerializerConfig
) -> _Element:
    if obj is None:
        return _make_element(element, xml_annotation.tag)

    # An optional is a union where the last element is the None type.
    union_types = typing_inspect.get_args(type_annotation)[:-1]
    if len(union_types) == 1:
        # This was Optional[T]
        return _from_obj(
            obj,
            union_types[0],
            xml_annotation,
            element,
            config
        )
    else:
        return _from_union(
            obj,
            Union[tuple(union_types)], # type: ignore
            xml_annotation,
            element,
            config
        )


def _from_union(
        obj: Any,
        type_annotation: Annotation,
        xml_annotation: XMLAnnotation,
        element: Optional[_Element],
        config: SerializerConfig
) -> _Element:
    for union_type_annotation in typing_inspect.get_args(type_annotation):
        try:
            return _from_obj(
                obj,
                union_type_annotation,
                xml_annotation,
                element,
                config
            )
        except:  # pylint: disable=bare-except
            pass

    raise ValueError('unable to find type that satisfies union')


def _from_list(
        obj: list,
        type_annotation: Annotation,
        xml_annotation: XMLAnnotation,
        element: Optional[_Element],
        config: SerializerConfig
) -> _Element:
    item_annotation, *_rest = typing_inspect.get_args(type_annotation)
    if typing_inspect.is_annotated_type(item_annotation):
        item_type_annotation, item_xml_annotation = get_xml_annotation(
            item_annotation,
            DEFAULT_VALUE_ANNOTATION
        )
    else:
        item_type_annotation = item_annotation
        item_xml_annotation = xml_annotation

    if element is None:
        element = Element(xml_annotation.tag)

    if xml_annotation.tag == item_xml_annotation.tag:
        # siblings
        parent = element
    else:
        parent = _make_element(element, xml_annotation.tag)

    for item in obj:
        _from_obj(
            item,
            item_type_annotation,
            item_xml_annotation,
            parent,
            config
        )

    return parent


def _from_typed_dict(
        obj: dict,
        type_annotation: Annotation,
        xml_annotation: XMLAnnotation,
        element: Optional[_Element],
        config: SerializerConfig
) -> _Element:
    dict_element = _make_element(element, xml_annotation.tag)

    typed_dict_keys = typing_inspect.typed_dict_keys(type_annotation)
    for key, key_annotation in typed_dict_keys.items():
        default = getattr(type_annotation, key, Parameter.empty)
        if typing_inspect.is_annotated_type(key_annotation):
            item_type_annotation, item_xml_annotation = get_xml_annotation(
                key_annotation,
                DEFAULT_KEY_ANNOTATION
            )
        else:
            tag = config.serialize_key(key) if isinstance(key, str) else key
            item_type_annotation = key_annotation
            item_xml_annotation = XMLEntity(tag)

        value = obj.get(key, default)
        if value is not Parameter.empty:
            _from_obj(
                value,
                item_type_annotation,
                item_xml_annotation,
                dict_element,
                config
            )
        else:
            # TODO: Should we throw?
            pass

    return dict_element


def _from_simple(
        obj: Any,
        type_annotation: Annotation,
        xml_annotation: XMLAnnotation,
        element: Optional[_Element],
        config: SerializerConfig
) -> _Element:
    text = _from_value(obj, type_annotation, config)
    if isinstance(xml_annotation, XMLAttribute):
        if element is None:
            raise ValueError("No element for attribute")
        element.set(xml_annotation.tag, text)
        return element
    else:
        child = _make_element(element, xml_annotation.tag)
        child.text = text
        return child


def _from_obj(
        obj: Any,
        type_annotation: Annotation,
        xml_annotation: XMLAnnotation,
        element: Optional[_Element],
        config: SerializerConfig
) -> _Element:
    if is_simple_type(type_annotation):
        return _from_simple(
            obj,
            type_annotation,
            xml_annotation,
            element,
            config
        )
    elif typing_inspect.is_optional_type(type_annotation):
        return _from_optional(
            obj,
            type_annotation,
            xml_annotation,
            element,
            config
        )
    elif typing_inspect.is_list_type(type_annotation):
        return _from_list(
            obj,
            type_annotation,
            xml_annotation,
            element,
            config
        )
    elif typing_inspect.is_typed_dict_type(type_annotation):
        return _from_typed_dict(
            obj,
            type_annotation,
            xml_annotation,
            element,
            config
        )
    elif typing_inspect.is_union_type(type_annotation):
        return _from_union(
            obj,
            type_annotation,
            xml_annotation,
            element,
            config
        )
    else:
        raise TypeError


def serialize(
        obj: Any,
        annotation: Annotation,
        config: SerializerConfig
) -> str:
    type_annotation, xml_annotation = get_xml_annotation(
        annotation,
        DEFAULT_DOCUMENT_ANNOTATION
    )
    if not isinstance(xml_annotation, XMLEntity):
        raise TypeError(
            "Expected the root value to have an XMLEntity annotation")

    element = _from_obj(obj, type_annotation, xml_annotation, None, config)
    buf: bytes = etree.tostring(
        element,
        pretty_print=config.pretty_print)  # pylint: disable=c-extension-no-member
    return buf.decode()



"""XML annotations"""

from typing import Tuple, cast

from ..types import Annotation
from ..custom_annotations import (
    SerializationAnnotation,
    is_any_serialization_annotation,
    get_all_serialization_annotations
)


class XMLAnnotation(SerializationAnnotation):
    """The base XML annotation class"""

    def __init__(self, tag: str):
        self.tag = tag


class XMLEntity(XMLAnnotation):
    """An XML entity"""

    def __repr__(self) -> str:
        return f'XMLEntity("tag={self.tag}")'


class XMLAttribute(XMLAnnotation):
    """An XML attribute"""

    def __repr__(self) -> str:
        return f'XMLAttribute(tag="{self.tag}")'

DEFAULT_DOCUMENT_ANNOTATION = XMLEntity('document')
DEFAULT_KEY_ANNOTATION = XMLEntity('key')
DEFAULT_VALUE_ANNOTATION = XMLEntity('value')


def is_xml_annotation(annotation: Annotation) -> bool:
    """Determine if the annotation is of type Annotation[T, XMLAnnotation]

    Args:
        annotation (Any): The annotation

    Returns:
        bool: True if the annotation is of type Annotation[T, XMLAnnotation],
            otherwise False
    """
    if not is_any_serialization_annotation(annotation):
        return False
    _, serialization_annotations = get_all_serialization_annotations(
        annotation
    )
    xml_annotations = [
        serialization_annotation
        for serialization_annotation in serialization_annotations
        if issubclass(type(serialization_annotation), XMLAnnotation)
    ]
    return len(xml_annotations) == 1


def get_xml_annotation(
        annotation: Annotation,
        default: XMLAnnotation
) -> Tuple[Annotation, XMLAnnotation]:
    """Gets the type T of Annotation[T, XMLAnnotation]

    Args:
        annotation (Any): The annotation
        default (XMLAnnotation): The default annotation.

    Returns:
        Tuple[Annotation, XMLAnnotation]: The type and the XML annotation
    """
    type_annotation, serialization_annotations = get_all_serialization_annotations(
        annotation
    )
    xml_annotations = [
        serialization_annotation
        for serialization_annotation in serialization_annotations
        if issubclass(type(serialization_annotation), XMLAnnotation)
    ]
    xml_annotation = xml_annotations[0] if xml_annotations else default
    return type_annotation, cast(XMLAnnotation, xml_annotation)



"""XML annotations"""

from typing import Tuple, cast

from ..types import Annotation
from ..custom_annotations import (
    SerializationAnnotation,
    is_any_serialization_annotation,
    get_all_serialization_annotations
)


class JSONAnnotation(SerializationAnnotation):
    """The base JSON annotation class"""


class JSONValue(JSONAnnotation):
    """A JSON property"""

    def __repr__(self) -> str:
        return 'JSONValue()'


class JSONProperty(JSONAnnotation):
    """A JSON property"""

    def __init__(self, tag: str):
        self.tag = tag

    def __repr__(self) -> str:
        return f'JSONProperty(tag="{self.tag}")'


def is_json_annotation(annotation: Annotation) -> bool:
    """Determine if the annotation is of type Annotation[T, JSONAnnotation]

    Args:
        annotation (Any): The annotation

    Returns:
        bool: True if the annotation is of type Annotation[T, JSONAnnotation],
            otherwise False
    """
    if not is_any_serialization_annotation(annotation):
        return False
    _, serialization_annotations = get_all_serialization_annotations(
        annotation
    )
    json_annotations = [
        serialization_annotation
        for serialization_annotation in serialization_annotations
        if issubclass(type(serialization_annotation), JSONAnnotation)
    ]
    return len(json_annotations) == 1


def get_json_annotation(annotation: Annotation) -> Tuple[Annotation, JSONAnnotation]:
    """Gets the type T of Annotation[T, JSONAnnotation]

    Args:
        annotation (Any): The annotation

    Returns:
        Tuple[Annotation, JSONAnnotation]: The type and the JSON annotation
    """
    type_annotation, serialization_annotations = get_all_serialization_annotations(
        annotation
    )
    json_annotations = [
        serialization_annotation
        for serialization_annotation in serialization_annotations
        if issubclass(type(serialization_annotation), JSONAnnotation)
    ]
    return type_annotation, cast(JSONAnnotation, json_annotations[0])



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['jetblack_serialization',
 'jetblack_serialization.json',
 'jetblack_serialization.xml']

package_data = \
{'': ['*']}

install_requires = \
['jetblack-iso8601>=1.0,<2.0',
 'lxml>=4.9,<5.0',
 'typing-extensions>=4,<5',
 'typing_inspect>=0.8,<0.9']

setup_kwargs = {
    'name': 'jetblack-serialization',
    'version': '2.2.0rc1',
    'description': 'Serialization for JSON and XML using typing',
    'long_description': '# jetblack-serialization\n\nSerialization for JSON and XML in Python using typing annotations\n(read the [docs](https://rob-blackbourn.github.io/jetblack-serialization/)).\n\n## Status\n\nIt has been tested with Python 3.7 used the `typing_extensions`\npackage for `TypedDict` and `Annotated`. In Python 3.8 the `TypedDict`\nclass is available in the standard `typing` package.\n\n## Installation\n\nThe package can be installed with pip.\n\n```bash\npip install jetblack-serialization\n```\n\n## Overview\n\nThe package adds support for type annotations when serializing or deserializing\nJSON or XML.\n\n\n### JSON\n\nGiven a typed dictionary:\n\n```python\nfrom datetime import datetime\nfrom typing import List, Optional, TypedDict, Union\n\nclass Book(TypedDict, total=False):\n    book_id: int\n    title: str\n    author: str\n    publication_date: datetime\n    keywords: List[str]\n    phrases: List[str]\n    age: Optional[Union[datetime, int]]\n    pages: Optional[int]\n```\n\n#### Serializing\n\nThis could be serialized to JSON as:\n\n```python\nfrom stringcase import camelcase, snakecase\nfrom jetblack_serialization import SerializerConfig\nfrom jetblack_serialization.json import serialize\n\nobj: Book = {\n    \'author\': \'Chairman Mao\',\n    \'book_id\': 42,\n    \'title\': \'Little Red Book\',\n    \'publication_date\': datetime(1973, 1, 1, 21, 52, 13),\n    \'keywords\': [\'Revolution\', \'Communism\'],\n    \'phrases\': [\n        \'Revolutionary wars are inevitable in class society\',\n        \'War is the continuation of politics\'\n    ],\n    \'age\': 24,\n}\ntext = serialize(\n    obj,\n    Book,\n    SerializerConfig(camelcase, snakecase, pretty_print=True)\n)\nprint(text)\n```\n\ngiving:\n\n```json\n{\n    "bookId": 42,\n    "title": "Little Red Book",\n    "author": "Chairman Mao",\n    "publicationDate": "1973-01-01T21:52:13.00Z",\n    "keywords": ["Revolution", "Communism"],\n    "phrases": ["Revolutionary wars are inevitable in class society", "War is the continuation of politics"],\n    "age": 24,\n    "pages": null\n}\n```\n\nNote the fields have been camel cased, and the publication date has been turned\ninto an ISO 8601 date.\n\n#### Deserializing\n\nWe can deserialize the data as follows:\n\n```python\nfrom stringcase import camelcase, snakecase\nfrom jetblack_serialization import SerializerConfig\nfrom jetblack_serialization.json import deserialize\n\ndct = deserialize(\n    text,\n    Annotated[Book, JSONValue()],\n    SerializerConfig(camelcase, snakecase)\n)\n```\n\n### XML\n\nThe XML version of the typed dictionary might look like this:\n\n```python\nfrom datetime import datetime\nfrom typing import List, Optional, TypedDict, Union\nfrom typing_extensions import Annotated\nfrom jetblack_serialization.xml import XMLEntity, XMLAttribute\n\nclass Book(TypedDict, total=False):\n    book_id: Annotated[int, XMLAttribute("bookId")]\n    title: str\n    author: str\n    publication_date: datetime\n    keywords: Annotated[List[Annotated[str, XMLEntity("Keyword")]], XMLEntity("Keywords")]\n    phrases: List[str]\n    age: Optional[Union[datetime, int]]\n    pages: Optional[int]\n```\n\nNote we have introduced some annotations to control the serialization.\nFor XML we have used pascal-case to serialized the keys and snake-case\nfor deserialization.\n\n#### Serializing\n\nTo serialize we need to provide the containing tag `Book`:\n\n```python\nfrom stringcase import pascalcase, snakecase\nfrom jetblack_serialization import SerializerConfig\nfrom jetblack_serialization.xml import serialize\n\nbook: Book = {\n    \'author\': \'Chairman Mao\',\n    \'book_id\': 42,\n    \'title\': \'Little Red Book\',\n    \'publication_date\': datetime(1973, 1, 1, 21, 52, 13),\n    \'keywords\': [\'Revolution\', \'Communism\'],\n    \'phrases\': [\n        \'Revolutionary wars are inevitable in class society\',\n        \'War is the continuation of politics\'\n    ],\n    \'age\': 24,\n    \'pages\': None\n}\ntext = serialize(\n    book,\n    Annotated[Book, XMLEntity("Book")],\n    SerializerConfig(pascalcase, snakecase)\n)\nprint(text)\n```\n\nProducing:\n\n```xml\n<Book bookId="42">\n    <Title>Little Red Book</Title>\n    <Author>Chairman Mao</Author>\n    <PublicationDate>1973-01-01T21:52:13.00Z</PublicationDate>\n    <Keywords>\n        <Keyword>Revolution</Keyword>\n        <Keyword>Communism</Keyword>\n    </Keywords>\n    <Phrase>Revolutionary wars are inevitable in class society</Phrase>\n    <Phrase>War is the continuation of politics</Phrase>\n    <Age>24</Age>\n</Book>\'\n```\n\nThe annotations are more elaborate here. However, much of the typed dictionary\nrequires no annotation.\n\nFirst we needed the outer document wrapper `XMLEntity("Book")`.\n\nNext we annotated the `book_id` to be an `XMLAttribute`.\n\nFinally we annotated the two lists differently. The `keywords` list used\na nested structure, which we indicated by giving the list a different\n`XMLEntity` tag to the list items. For the phrases we used the default\nin-line behaviour.\n\n#### Deserializing\n\nWe can deserialize the XML as follows:\n\n```python\nfrom stringcase import pascalcase, snakecase\nfrom jetblack_serialization import SerializerConfig\nfrom jetblack_serialization.xml import deserialize\n\ndct = deserialize(\n    text,\n    Annotated[Book, XMLEntity("Book")],\n    SerializerConfig(pascalcase, snakecase)\n)\n```\n\n## Attributes\n\nFor JSON, attributes are typically not required. However\n`JSONProperty(tag: str)` and `JSONValue()` are provided for\ncompleteness.',
    'author': 'Rob Blackbourn',
    'author_email': 'rob.blackbourn@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/rob-blackbourn/jetblack-serialization',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



"""An untyped deserializer"""

import json
from typing import Any

from ..config import SerializerConfig


def _deserialize_key_if_str(key: Any, config: SerializerConfig) -> Any:
    return config.deserialize_key(
        key
    ) if config.deserialize_key and isinstance(key, str) else key


def _from_value(
        value: Any,
        config: SerializerConfig
) -> Any:
    if isinstance(value, str):
        for deserializer in config.value_deserializers.values():
            try:
                return deserializer(value)
            except:  # pylint: disable=bare-except
                pass
    return value


def _from_list(lst: list, config: SerializerConfig) -> list:
    return [
        _from_obj(item, config)
        for item in lst
    ]


def _from_dict(dct: dict, config: SerializerConfig) -> dict:
    return {
        _deserialize_key_if_str(key, config): _from_obj(value, config)
        for key, value in dct.items()
    }


def _from_obj(obj: Any, config: SerializerConfig) -> Any:
    if isinstance(obj, dict):
        return _from_dict(obj, config)
    elif isinstance(obj, list):
        return _from_list(obj, config)
    else:
        return _from_value(obj, config)


def deserialize(text: str, config: SerializerConfig) -> Any:
    """Deserialize JSON without type information

    Args:
        text (str): The JSON string

    Returns:
        Any: The deserialized JSON object
    """

    def _hook(obj: dict) -> dict:
        return _from_dict(obj, config)

    return json.loads(text, object_hook=_hook)



"""Extensions for typing_inspect"""

try:
    # Python3.8
    from typing import _TypedDictMeta  # type: ignore
except:  # pylint: disable=bare-except
    # Python3.7
    from typing_extensions import _TypedDictMeta  # type: ignore
# type: ignore# pylint: disable=unused-import
from typing_extensions import _AnnotatedAlias
from typing_inspect import (
    get_args,
    get_bound,
    get_constraints,
    get_generic_bases,
    get_generic_type,
    get_last_args,
    get_last_origin,
    get_origin,
    get_parameters,
    is_callable_type,
    is_classvar,
    is_generic_type,
    is_literal_type,
    # is_optional_type,
    is_tuple_type,
    is_typevar,
    is_union_type
)


def is_typed_dict_type(tp):
    """Return True if tp is a typed dict"""
    return isinstance(tp, _TypedDictMeta)


def is_list_type(tp):
    """Return True if tp is a list"""
    return (
        get_origin(tp) is list
        and getattr(tp, '_name', None) == 'List'
    )


def is_annotated_type(tp):
    return isinstance(tp, _AnnotatedAlias)


def get_metadata(annotation):
    return getattr(annotation, '__metadata__', None)


def get_optional_type(annotation):
    """Get the nested type annotation T for an Optional[T]"""
    if not is_optional_type(annotation):
        return None
    contained_type, *_rest = get_args(annotation)
    return contained_type


def is_dict_type(tp):
    """Return True if tp is a Dict"""
    return (
        get_origin(tp) is dict
        and getattr(tp, '_name', None) == 'Dict'
    )


def typed_dict_keys(td):
    """If td is a TypedDict class, return a dictionary mapping the typed keys to types.
    Otherwise, return None. Examples::

        class TD(TypedDict):
            x: int
            y: int
        class Other(dict):
            x: int
            y: int

        typed_dict_keys(TD) == {'x': int, 'y': int}
        typed_dict_keys(dict) == None
        typed_dict_keys(Other) == None
    """
    if isinstance(td, _TypedDictMeta):
        return td.__annotations__.copy()
    return None


def get_unannotated_type(tp):
    while is_annotated_type(tp):
        tp = get_origin(tp)
    return tp


def is_optional_type(tp):
    """Test if the type is type(None), or is a direct union with it, such as Optional[T].

    NOTE: this method inspects nested `Union` arguments but not `TypeVar` definition
    bounds and constraints. So it will return `False` if
     - `tp` is a `TypeVar` bound, or constrained to, an optional type
     - `tp` is a `Union` to a `TypeVar` bound or constrained to an optional type,
     - `tp` refers to a *nested* `Union` containing an optional type or one of the above.

    Users wishing to check for optionality in types relying on type variables might wish
    to use this method in combination with `get_constraints` and `get_bound`
    """

    if tp is type(None):  # noqa
        return True
    elif is_union_type(tp):
        return any(is_optional_type(tt) for tt in get_args(tp, evaluate=True))
    elif is_annotated_type(tp):
        return is_optional_type(get_unannotated_type(tp))
    else:
        return False
